/*

IIASA blue + three-stage gradient to white
- #00589d / #3f81b5 / #7fabce / #bfd5e6

IIASA secondary colours + two-stage gradient to white
- #62c5c0 / #96d8d5 / #caebea
- #247f6e
- #fdbb40
- #ee6a6b
- #6a4c93

*/

/* Overall backgrounds */
.wy-nav-side, .wy-side-nav-search, .wy-nav-top {
  background: #00589d;
}

/* White line between search & menu panels */
.wy-side-nav-search {
  border-bottom: 1px solid #ffffff;
}

/* Text colours of menu items: general */
.wy-menu-vertical a {
  color: #ffffff;
}

/* …: current items levels 1–3 */
.wy-menu-vertical li.on a,
.wy-menu-vertical li.current > a,
.wy-menu-vertical li.toctree-l3.current > a {
  color: #00589d;
}

/* …: current item level 4 */
.wy-menu-vertical li.toctree-l3 > a {
  color: #000000;
}

/* Borders around current items */
.wy-menu-vertical li.current a {
  border-right: 1px solid #00589d;
}

.wy-menu-vertical li.toctree-l1.current > a {
  border-bottom: 1px solid #00589d;
  border-right: none;
}

/* Gradient colouring of sub-menus */
.wy-menu-vertical li.current,
.wy-menu-vertical li.toctree-l2.current > a {
  background: #bfd5e6; /* blue */
}


.wy-menu-vertical li.toctree-l2.current li.toctree-l3 > a {
  background: #7fabce; /* blue */
}

.wy-menu-vertical li.toctree-l3.current li.toctree-l4 > a {
  background: #3f81b5; /* blue */
  color: #ffffff;
}

/* Hovered links; all the same colour */
.wy-menu-vertical li > a:hover,
.wy-menu-vertical li.current > a:hover,
.wy-menu-vertical li.current li.toctree-l2 > a:hover,
.wy-menu-vertical li.toctree-l2.current li.toctree-l3 > a:hover,
.wy-menu-vertical li.toctree-l3.current li.toctree-l4 > a:hover {
  /* white on purple */
  background: #6a4c93;
  color: #ffffff;
}
